/**
 * Title: GenerateCertificates
 *
 * Description: Sample application to generate security certificates, for use
 * with the secure CORBA sample applications.
 *
 * Copyright: Copyright (c) 2004
 *
 * Company: MetaSwitch
 *
 * @author Thomas Price and Alex Wilber, EMS Team
 *
 * @version 1.0
 */

//-----------------------------------------------------------------------------
// This application generates two key pairs, and uses them to create two
// certificates - a client certificate, and a Certificate Authority (CA)
// certificate.  The CA certificate is then used to sign the client
// certificate.
//-----------------------------------------------------------------------------
import java.io.*;
import java.util.*;
import java.math.*;
import java.security.*;
import iaik.security.provider.*;
import iaik.asn1.*;
import iaik.asn1.structures.*;
import iaik.x509.*;
import iaik.pkcs.pkcs8.*;

public class GenerateCertificates
{
  static private KeyPair sCAKeyPair;
  static private KeyPair sClientKeyPair;

  //---------------------------------------------------------------------------
  // Variables to store the command line input
  //---------------------------------------------------------------------------
  static private String sCountry;
  static private String sLocality;
  static private String sOrganization;
  static private String sEmail;
  static private String sCACommonName;
  static private String sCAPrefix;
  static private String sClientCommonName;
  static private String sClientPrefix;

  public static void main(String[] args)
  {
    //-------------------------------------------------------------------------
    // Check that the correct number of arguments has been received.  If it
    // hasn't, output a description of the proper usage and exit.
    //-------------------------------------------------------------------------
    if (args.length != 8)
    {
      System.err.println("Usage: GenerateCertificates Country Locality "
                       + "Organization\n"
                       + "       EmailAddress CACommonName "
                       + "CACertificateFileNamePrefix\n"
                       + "       ClientCommonName "
                       + "ClientCertificateFileNamePrefix");

      System.exit(1);
    }

    //-------------------------------------------------------------------------
    // Set the values of static variables using the command line arguments.
    //-------------------------------------------------------------------------
    sCountry = args[0];
    sLocality = args[1];
    sOrganization = args[2];
    sEmail = args[3];
    sCACommonName = args[4];
    sCAPrefix = args[5];
    sClientCommonName = args[6];
    sClientPrefix = args[7];

    //-------------------------------------------------------------------------
    // Register IAIK as a security provider.
    //-------------------------------------------------------------------------
    IAIK.addAsProvider(true);

    try
    {
      //-----------------------------------------------------------------------
      // Create a KeyPairGenerator object and use it to generate key pairs for
      // the Certificate Authority and the Client.
      //-----------------------------------------------------------------------
      KeyPairGenerator generator = KeyPairGenerator.getInstance("RSA", "IAIK");

      System.out.println("Generating CA keypair...");

      sCAKeyPair = generator.generateKeyPair();
      saveKey(sCAKeyPair, sCAPrefix, "EmsEms");

      System.out.println("Generating Client keypair...");

      sClientKeyPair = generator.generateKeyPair();
      saveKey(sClientKeyPair, sClientPrefix, "EmsEms");

      //-----------------------------------------------------------------------
      // Create the certificates
      //-----------------------------------------------------------------------
      createCertificates();

      System.out.println("Done");
    }
    //-------------------------------------------------------------------------
    // Catch any unexpected exceptions.
    //-------------------------------------------------------------------------
    catch(NoSuchAlgorithmException ex)
    {
      System.err.println(ex.getMessage());
    }
    catch(NoSuchProviderException ex)
    {
      System.err.println(ex.getMessage());
    }

    System.exit(0);
  }

  /**
   * Outputs a security key to a file.
   *
   * @param keyPair     The keypair containing the key to save.
   * @param fileNamePrefix
   *                    The key will be saved in a .key file with this name.
   * @param passPhrase
   *                    The passphrase to use.
   */
  private static void saveKey(KeyPair keyPair,
                              String fileNamePrefix,
                              String passPhrase)
  {
    try
    {
      //-----------------------------------------------------------------------
      // The name of the key file depends on the parameter 'fileNamePrefix',
      // which in turn depends on a command line argument.
      //-----------------------------------------------------------------------
      String fileName = fileNamePrefix + ".key";
      System.out.println("saving " + fileNamePrefix + " encrypted private key "
                                                           + "in " + fileName);

      //-----------------------------------------------------------------------
      // Encrypt the key.
      //-----------------------------------------------------------------------
      PrivateKey securityKey = keyPair.getPrivate();
      EncryptedPrivateKeyInfo epki = new EncryptedPrivateKeyInfo(securityKey);
      epki.encrypt(passPhrase,
                   AlgorithmID.pbeWithMD5AndDES_CBC,
                   new SecureRandom());

      //-----------------------------------------------------------------------
      // Write the file.
      //-----------------------------------------------------------------------
      FileOutputStream fos = new FileOutputStream(fileName);
      fos.write(epki.getEncoded());
      fos.close();
    }
    //-------------------------------------------------------------------------
    // Catch any unexpected exceptions.
    //-------------------------------------------------------------------------
    catch(NoSuchAlgorithmException ex)
    {
      System.err.println(ex.getMessage());
    }
    catch(IOException ex)
    {
      System.err.println(ex.getMessage());
    }
  }

  /**
   * Create the certificate files.
   */
  public static void createCertificates()
  {
    //-------------------------------------------------------------------------
    // Build the certificate for the Certificate Authority.  Pass command line
    // details to a Name object, and use it to create the certificate.  Here
    // the Certificate Authority (CA) key pair is used both to create and to
    // sign the certificate, i.e.  the certificate is self-signed.
    //-------------------------------------------------------------------------
    System.out.println("Creating self-signed CA certificate for " +
                                                        sCACommonName + "...");

    Name caName = new Name();

    caName.addRDN(ObjectID.country, sCountry);
    caName.addRDN(ObjectID.locality, sLocality);
    caName.addRDN(ObjectID.organization, sOrganization);
    caName.addRDN(ObjectID.commonName, sCACommonName);
    caName.addRDN(ObjectID.emailAddress, sEmail);

    X509Certificate caCert = createCertificate(caName,
                                               sCAKeyPair,
                                               caName,
                                               sCAKeyPair);

    //-------------------------------------------------------------------------
    // Save the CA certificate using the prefix given as a command line
    // argument.
    //-------------------------------------------------------------------------
    saveCertificate(caCert, sCAPrefix);

    //-------------------------------------------------------------------------
    // Build the certificate for the client.  Again, pass command line details
    // to a Name object, and use it to build the certificate.  Here the
    // certificate is created using the Client keypair, but it is signed using
    // the CA's keypair.
    //-------------------------------------------------------------------------
    System.out.println("Creating client certificate for " +
                                                    sClientCommonName + "...");

    Name clientName = new Name();

    clientName.addRDN(ObjectID.country, sCountry);
    clientName.addRDN(ObjectID.locality, sLocality);
    clientName.addRDN(ObjectID.organization, sOrganization);
    clientName.addRDN(ObjectID.commonName, sClientCommonName);
    clientName.addRDN(ObjectID.emailAddress, sEmail);

    X509Certificate clientCert = createCertificate(clientName,
                                                   sClientKeyPair,
                                                   caName,
                                                   sCAKeyPair);

    //-------------------------------------------------------------------------
    // Save the CA certificate using the prefix given as a command line
    // argument.
    //-------------------------------------------------------------------------
    saveCertificate(clientCert, sClientPrefix);
  }

  /**
   * Create an individual certificate.
   *
   * @param subject     The name of the subject to create a certificate for.
   * @param subjectKeyPair
   *                    The subject's key pair.
   * @param issuer      The name of the certificate issuer.
   * @param isuerKeyPair
   *                    The issuer's key pair.
   */
  private static X509Certificate createCertificate(Name subject,
                                                   KeyPair subjectKeyPair,
                                                   Name issuer,
                                                   KeyPair issuerKeyPair)
  {
    PublicKey publicKey = subjectKeyPair.getPublic();
    PrivateKey securityKey = issuerKeyPair.getPrivate();

    X509Certificate result = null;

    try
    {
      //-----------------------------------------------------------------------
      // Create a new certificate object and set its properties using the
      // method's parameters
      //-----------------------------------------------------------------------
      result = new X509Certificate();

      result.setSerialNumber(BigInteger.valueOf(0));
      result.setSubjectDN(subject);
      result.setPublicKey(publicKey);
      result.setIssuerDN(issuer);

      //-----------------------------------------------------------------------
      // Set the time properties for the certificate.  It is valid for 1000
      // years.
      //-----------------------------------------------------------------------
      GregorianCalendar date = new GregorianCalendar();
      date.setTimeInMillis(0);
      result.setValidNotBefore(date.getTime());
      date.add(Calendar.YEAR, 1000);
      result.setValidNotAfter(date.getTime());

      //-----------------------------------------------------------------------
      // Sign the certificate.
      //-----------------------------------------------------------------------
      result.sign(AlgorithmID.md5WithRSAEncryption, securityKey);
    }
    //-------------------------------------------------------------------------
    // Catch any unexpected exceptions.
    //-------------------------------------------------------------------------
    catch(java.security.InvalidKeyException ex)
    {
      System.err.println(ex.getMessage());
    }
    catch(java.security.NoSuchAlgorithmException ex)
    {
      System.err.println(ex.getMessage());
    }
    catch(java.security.cert.CertificateException ex)
    {
      System.err.println(ex.getMessage());
    }

    return result;
  }

  /**
   * Save a certificate to a file.
   *
   * @param certificate
   *                    The certificate to output to a file.
   * @param fileNamePrefix
   *                    The certificate will be saved in a .der file with this
   *                    name.
   */
  private static void saveCertificate(X509Certificate certificate,
                                      String fileNamePrefix)
  {
    //-------------------------------------------------------------------------
    // Save the certificate to a file.  The filename used depends on the
    // parameter 'fileNamePrefix', which in turn depends on a command line
    // argument.
    //-------------------------------------------------------------------------
    String fileName = fileNamePrefix + ".der";
    System.out.println("Saving " + fileNamePrefix + " certificate in "
                                                                   + fileName);
    try
    {
      //-----------------------------------------------------------------------
      // Create a FileOutputStream object and use it to write the file.
      //-----------------------------------------------------------------------
      FileOutputStream fos = new FileOutputStream(fileName);
      fos.write(certificate.getEncoded());
      fos.close();
    }
    //-------------------------------------------------------------------------
    // Catch any unexpected exceptions.
    //-------------------------------------------------------------------------
    catch(IOException ex)
    {
      System.err.println(ex.getMessage());
    }
    catch(java.security.cert.CertificateEncodingException ex)
    {
      System.err.println(ex.getMessage());
    }
  }
}
